#!/bin/bash
#set -x
#Globals
MAPR_HOME="${MAPR_HOME:-__PREFIX__}"
drillBase="${MAPR_HOME}/drill"
drill_secure="false"
drill_qs_only=0
drill_only=0
drill_qs_and_drill=0
zkNodesList=""
drillQSEnabled=0
drillBitsNodeOn=0
drillQSNodeOn=0
wardenTmpDrillConf="/tmp/warden.drill-bits.conf"
wardenInstalledDrillConf="${MAPR_CONF_DIR}/warden.drill-bits.conf"
wardenConfFileInstalled=0
httpUserName=""
httpUserPwd=""
httpPort="8047"
httpsLoginResult=""
httpsCookieFile="/tmp/installer_cookie.txt"
DRILL_CONF_ASSUME_RUNNING_CORE=1

# Storage Plugin related parameters
hiveMetastoreHost=$HIVEMETASTORE_HOST
hiveMetastoreDbHost=$HIVEMETASTORE_DATABASE_HOST

# Security related parameters
drillClientAuthMech="[\"MAPRSASL\", \"PLAIN\"]"
drillBitAuthMech="MAPRSASL"
drillAuthPackage="org.apache.drill.exec.rpc.user.security"
drillAuthImpl="pam"
drillPamProfiles="[\"sudo\", \"login\", \"mapr-admin\"]"

# Security configuration prefixes
impersonationPrefix=impersonation
securityUserPrefix=security.user
securityUserAuthPrefix=$securityUserPrefix.auth
securityBitPrefix=security.bit
httpsPrefix=http

if [ -e "${MAPR_HOME}/server/common-ecosystem.sh" ]; then
    . "${MAPR_HOME}/server/common-ecosystem.sh"
else
   echo "Failed to source common-ecosystem.sh"
   exit 0
fi


function GetDrillVersion() {
  local ver=""
  if [ -f "${drillBase}/drillversion" ]; then
    ver=$(cat "${drillBase}/drillversion")
  else
    ver=`ls -t ${drillBase} | sed 's/^.*drill-//' | head -1 | awk '{print $1}'`
  fi
  echo "$ver"
}

function ConfigureRunUserForDrill() {
  local DRILL_DIR CURR_USER CURR_GROUP
  DRILL_DIR=$1
  if [ ! -d $DRILL_DIR/conf ]; then
    return
  fi

  CURR_USER=$2
  CURR_GROUP=$3
  chown -R $CURR_USER:$CURR_GROUP "${DRILL_DIR}/conf" >> $logFile 2>&1
}

function getMemValueFromWardenFile() {
    local rc
    local wFile=$1
    local wIdent=$2
    local wVariable=$3
    local RES=$(cat $wFile | grep "^$wIdent" | sed -e 's/.*'"$wVariable="'\([0-9]*[GgMmKk]*\).*$/\1/')
    rc=$?
    echo "$RES"
    return $rc
}

# Copy libjpam out of /opt/mapr/lib directory inside DRILL_HOME/conf/pam directory.
# Then use this path inside distrib-env.sh for java.library.path. We want to avoid
# directly pointing to /opt/mapr/lib since it contains lots of other native libraries
# and jars which can conflict with Drill's version of same libs/jars
# 
# Only called in case when security is enabled
#
function CopyAndUpdatePamLib() {
    # Copy the pam library from $MAPR_HOME/lib to $DRILL_HOME/conf/pam
    local libPamBase="$MAPR_HOME/lib"
    local libPamSrc="$libPamBase/libjpam.so"
    
    if [ ! -f "$libPamSrc" ]
    then
        logInfo "ERROR: Failed to find libjpam.so under $libPamSrc path"
	return
    fi
    
    # If we are here that means file is present so copy it inside the DRILL_HOME/conf/path
    local drillBaseDir=${drillBase}/drill-${DRILL_VER}
    local drillConfDir="$drillBaseDir/conf"
    local libPamDest="$drillConfDir/pam"

    if [! -d "$drillConfDir" ]
    then
        logInfo "ERROR: Failed to find Drill's conf path: $drillConfDir"
	return
    fi

    # Remove and create pam directory
    rm -rf $libPamDest
    mkdir $libPamDest

    if [ $? -ne 0 ]
    then
        logInfo "ERROR: Failed to create pam directory: $libPamDest"
        return
    fi

    # Copy the libjpam.so
    cp -f $libPamSrc $libPamDest
    
    if [ $? -ne 0 ]
    then
        logInfo "ERROR: Failed to copy the libjpam under $libPamDest"
	return
    fi

    # Update distrib-env.sh to correct path of libjpam
    UpdatePamPath true DRILL_JAVA_OPTS $libPamDest
}

# Function to update the PAM library path in distrib-env.sh file based on security is enabled
# or disabled.
#
# In this function ~ is chosen as deliminiter for sed purposefully since pamPath has / in its value
# which causes invalid sed expression to form. 
#
# Parameters:
#           $1 = true/false -- security is enabled/disabled
#           $2 = DRILL_JAVA_OPTS
#           $3 = path of libjpam.so
#
function UpdatePamPath() {
     local javaLibPath="-Djava.library.path="

     # Seucrity is disabled hence remove the java.library.path from DRILL_JAVA_OPTS
     if [ "$1" == "false" ]
     then
         sed -i -e "s~ $javaLibPath[^ \"]*~~g" $drillDistribEnv
         return
     fi
     
     # Check if this string exists in distrib-env.sh
     if grep -q "$2.*$javaLibPath" $drillDistribEnv
     then
         # Remove the java.library.path pointing to old pam module path
	 sed -i -e "s~ $javaLibPath[^ \"]*~~g" $drillDistribEnv
     fi
     
     # Remove the last " in the DRILL_JAVA_OPTS
     sed -i -e "s~\(^export $2=.*\)\"$~\1~g" $drillDistribEnv

     # Add java.library.path in DRILL_JAVA_OPTS with pam module path
     sed -i -e "s~\(^export $2=.*\)~\1 $javaLibPath$3\"~g" $drillDistribEnv
}

# Add or Update zkSaslClientString as true/false for DRILL_JAVA_OPTS
# and SQLLINE_JAVA_OPTS variable inside distrib-env.sh file
#
# Parameters:
#           $1 = DRILL_JAVA_OPTS / SQLLINE_JAVA_OPTS
#           $2 = true/false -- value to change from
#           $3 = true/false -- value to change to
#
function UpdateZookeeperAuth() { 
    # Update the distrib-env.sh file to enable zookeeper security.
    local zkSaslClientString="zookeeper.sasl.client"

    # Check if this string exists in the distrib-env.sh
    if grep -q "$1.*$zkSaslClientString" $drillDistribEnv
    then
        # Update the value to true if it's set to $3 from $2 for all occurences
        sed -i -e "s/"$zkSaslClientString=$2"/"$zkSaslClientString=$3"/g" $drillDistribEnv
    else
        # The zkSaslClientString is absent we need to add one to the end
        # Remove the last " from $1
        sed -i -e "s/\(^export $1=.*\)\"$/\1/g" $drillDistribEnv

        # Now add the zkSaslClientString with " in the end
        sed -i -e "s/\(^export $1=.*\)/\1 -D$zkSaslClientString=$3\"/g" $drillDistribEnv
    fi
}

# Add impersonation config inside the drill-distrib.conf file for Drill
#
# It expects drill-distrib.conf file to be in correct state such that just adding the
# impersonation config string will be done inside drill.exec block by default
#
function AddImpersonationConfig() {
    echo -e "\t$impersonationPrefix.enabled: true," >> $drillDistribConf
    echo -e "\t$impersonationPrefix.max_chained_user_hops: 3,\n" >> $drillDistribConf
}

# Add or Update security configuration inside distrib.conf file for Drill
#
# Enable Authentication/Encryption on Drill
# Enable HTTPS for web channel
#
# Parameter:
#	$1: true/false -- value of enabled: field in security parameters.
#
function EnableSecurityInDistribConf() {
    
    # Remove the last closing brace from the config file
    sed -i -e "/}$/d" $drillDistribConf

    # Add config to enable ZK ACL's
    echo -e "\tzk.apply_secure_acl: $1,\n" >> $drillDistribConf

    # Add client related security config
    AddImpersonationConfig

    echo -e "\tsecurity.auth.mechanisms: $drillClientAuthMech," >> $drillDistribConf
    echo -e "\t$securityUserAuthPrefix.enabled: $1," >> $drillDistribConf
    echo -e "\t$securityUserAuthPrefix.packages += \"$drillAuthPackage\"," >> $drillDistribConf
    echo -e "\t$securityUserAuthPrefix.impl: \"$drillAuthImpl\"," >> $drillDistribConf
    echo -e "\t$securityUserAuthPrefix.pam_profiles: $drillPamProfiles," >> $drillDistribConf
    echo -e "\t$securityUserPrefix.encryption.sasl.enabled: $1," >> $drillDistribConf
    
    # Add server related security config
    echo -e "\n\t$securityBitPrefix.auth.enabled: $1," >> $drillDistribConf
    echo -e "\t$securityBitPrefix.auth.mechanism: \"$drillBitAuthMech\"," >> $drillDistribConf
    echo -e "\t$securityBitPrefix.encryption.sasl.enabled: $1," >> $drillDistribConf

    # Add config to enable ssl
    echo -e "\n\t$httpsPrefix.ssl_enabled: $1," >> $drillDistribConf

    # Add closing braces in the end
    echo -e "\n}" >> $drillDistribConf
}

# Delete all the security related configuration added while enabling the security
# for drillbit
function RemoveDrillSecurityConfig() {
    # Remove any blank lines
    sed -i -e "/^$/d" $drillDistribConf

    # Remove ZK ACL's related configs
    sed -i -e "/zk.apply_secure_acl.*/d" $drillDistribConf

    # Remove impersonation related configs
    sed -i -e "/$impersonationPrefix.*/d" $drillDistribConf

    # Remove drill client related security configs
    sed -i -e "/security.auth.*/d" $drillDistribConf
    sed -i -e "/$securityUserPrefix.*/d" $drillDistribConf

    # Remove drill bit related security configs
    sed -i -e "/$securityBitPrefix.*/d" $drillDistribConf

    # Remove HTTPS ssl configuration
    sed -i -e "/$httpsPrefix.*/d" $drillDistribConf
}

# Enable security for drill
function EnableSecurity() {

    # Update DRILL_JAVA_OPTS inside distrib-env.sh file
    UpdateZookeeperAuth DRILL_JAVA_OPTS false true
   
    # Update SQQLINE_JAVA_OPTS inside distrib-env.sh file
    UpdateZookeeperAuth SQLLINE_JAVA_OPTS false true

    # Copy and update the libjpam.so path inside distrib-env.sh file
    CopyAndUpdatePamLib

    # Remove drill security config from distrib.conf
    RemoveDrillSecurityConfig
    
    # Update distrib.conf file with security enabled
    EnableSecurityInDistribConf true
}

# Disable security for drill
function DisableSecurity() {

    # Update DRILL_JAVA_OPTS inside distrib-env.sh file
    UpdateZookeeperAuth DRILL_JAVA_OPTS true false

    # Update SQQLINE_JAVA_OPTS inside distrib-env.sh file
    UpdateZookeeperAuth SQLLINE_JAVA_OPTS true false

    # Update the Pam lib path inside distrib-env.sh file
    UpdatePamPath false DRILL_JAVA_OPTS

    # Remove drill security config from distrib.conf
    RemoveDrillSecurityConfig

    # If configuration is done for Drill as a QS only or for both QS and Drill
    # We will add impersonation setting in insecure case too
    if [ "$drill_qs_only" -eq 1 -o "$drill_qs_and_drill" -eq 1 ]; then
        # Remove the last braces in config file
        sed -i -e "/}$/d" $drillDistribConf

        # Add Impersonation settings
        AddImpersonationConfig

        # Add last braces in config file
        echo -e "\n}" >> $drillDistribConf
    fi
}

function ConfigureDrillBitsRole() {
  local numThreads=4
  local heapMem=""
  local rpcStanza="\n"
  local oldClusterId="drillbits1"
  local oldZkConnect="localhost:2181"
  local newClusterId="${clusterName}-drillbits"


  if ! hasRole 'drill-bits' ; then
    logInfo "Skipping Drill Bits Role configuration... Not found"
    return
  else
    drillBitsNodeOn=1
  fi

  logInfo "Configuring Drill Bits Role"
  local DRILL_BASE_DIR=${drillBase}/drill-${DRILL_VER}

  ConfigureRunUserForDrill $DRILL_BASE_DIR $MAPR_USER $MAPR_GROUP

  local DRILL_CONF_DIR="${DRILL_BASE_DIR}/conf"
  local drillConf="${DRILL_CONF_DIR}/drill-override.conf"
  local drillDistribEnv="${DRILL_CONF_DIR}/distrib-env.sh"
  local drillDistribConf="${DRILL_CONF_DIR}/drill-distrib.conf"
  local drillDistribQSMemConf="${DRILL_CONF_DIR}/drill-distrib-mem-qs.conf"

  if [ -e "$drillDistribConf" -a "$drillQSEnabled" -eq 0 ]; then
    if hasRole 'drill-qs' ; then
       # -QS option was already used - keep using it
       drillQSEnabled=1
       drillQSNodeOn=1
    fi
  fi

  if grep -q "cluster-id: \"$oldClusterId\"" $drillConf; then
    # we have a new install with default information - remove it
    sed -i -e "/cluster-id: \"$oldClusterId\"/d" $drillConf
    sed -i -e "/zk.connect: \"$oldZkConnect\"/d" $drillConf
  fi

  # always configure the number of threads required for QS
  if [ -e "$drillDistribQSMemConf" ]; then
    numThreads=$(grep DrillRpcUserThreads "$drillDistribQSMemConf" | cut -d'=' -f2)
    if [ -z "$numThreads" ]; then
      numThreads="4"
    fi
  fi
  rpcStanza="\n\trpc {\n\t\tuser.client.threads: $numThreads\n\t}\n"

  if [ "$drillQSEnabled" -eq 1 ]; then
    touch "${MAPR_HOME}/roles/drill-qs"
    if [ -e "$drillDistribQSMemConf" ]; then
      heapMem=$(grep DrillQSWardenHeapTotal "$drillDistribQSMemConf" | cut -d'=' -f2)
    fi
  fi

  # TODO improve this to remove drill.exec stanza only instead of overwriting file
  echo -e "drill.exec {\n\tcluster-id: \"$newClusterId\",\n\tzk.connect: \"$zkNodesList\"${rpcStanza}}\n" > $drillDistribConf

  drillEnvExistingExport=`grep "export HADOOP_HOME" ${drillDistribEnv}`
  if [ -z "$drillEnvExistingExport" ]; then
    echo "" >> $drillDistribEnv
    echo "export HADOOP_HOME=/opt/mapr/hadoop/hadoop-${hadoopVersion}" >> $drillDistribEnv
  else
    sed -i -e "s/^export HADOOP_HOME=.*/export HADOOP_HOME=\/opt\/mapr\/hadoop\/hadoop-${hadoopVersion}/g" $drillDistribEnv
  fi

  local wardenPkgDrillConf="${DRILL_CONF_DIR}/warden.drill-bits.conf"
  local conf_drill_heap=$(getMemValueFromWardenFile $wardenConfFile service.env DRILL_HEAP)
  local conf_drill_mdm=$(getMemValueFromWardenFile $wardenConfFile service.env DRILL_MAX_DIRECT_MEMORY)
  local conf_drill_dccs=$(getMemValueFromWardenFile $wardenConfFile service.env DRILLBIT_CODE_CACHE_SIZE)
  local conf_drill_heap_min=$(getMemValueFromWardenFile $wardenConfFile service.heapsize.min service.heapsize.min)
  local conf_drill_heap_max=$(getMemValueFromWardenFile $wardenConfFile service.heapsize.max service.heapsize.max)
  local def_drill_heap=$(getMemValueFromWardenFile $wardenPkgDrillConf service.env DRILL_HEAP)
  local def_drill_mdm=$(getMemValueFromWardenFile $wardenPkgDrillConf service.env DRILL_MAX_DIRECT_MEMORY)
  local def_drill_dccs=$(getMemValueFromWardenFile $wardenPkgDrillConf service.env DRILLBIT_CODE_CACHE_SIZE)
  local def_drill_heap_min=$(getMemValueFromWardenFile $wardenPkgDrillConf service.heapsize.min service.heapsize.min)
  local def_drill_heap_max=$(getMemValueFromWardenFile $wardenPkgDrillConf service.heapsize.max service.heapsize.max)
  local def_qs_heap=$(getMemValueFromWardenFile $drillDistribQSMemConf 'export DRILL_HEAP' DRILL_HEAP)
  local def_qs_mdm=$(getMemValueFromWardenFile $drillDistribQSMemConf 'export DRILL_MAX_DIRECT_MEMORY' DRILL_MAX_DIRECT_MEMORY)
  local def_qs_dccs=$(getMemValueFromWardenFile $drillDistribQSMemConf 'export DRILLBIT_CODE_CACHE_SIZE' DRILLBIT_CODE_CACHE_SIZE)
  local def_qs_heap_min=$(getMemValueFromWardenFile $drillDistribQSMemConf 'DrillQSWardenHeapTotal' DrillQSWardenHeapTotal)
  local def_qs_heap_max=$def_qs_heap_min
  local new_drill_heap=""
  local new_drill_mdm=""
  local new_drill_dccs=""
  local new_drill_heap_min=""
  local new_drill_heap_max=""
  local updateMemSettings=0

  if [ -e "$wardenInstalledDrillConf" ]; then
    cp "$wardenInstalledDrillConf" "$wardenTmpDrillConf"
    wardenConfFileInstalled=1
  else
    cp "$wardenPkgDrillConf" "$wardenTmpDrillConf"
  fi
  wardenConfFile="$wardenTmpDrillConf"
  # configure memory requirement
  if [ "$drillQSEnabled" -eq 0 ]; then
      new_drill_heap=$def_drill_heap
      new_drill_mdm=$def_drill_mdm
      new_drill_dccs=$def_drill_dccs
      new_drill_heap_min=$def_drill_heap_min
      new_drill_heap_max=$def_drill_heap_max
  else
      new_drill_heap="${def_qs_heap}m"
      new_drill_mdm="${def_qs_mdm}m"
      new_drill_dccs="${def_qs_dccs}m"
      new_drill_heap_min="$def_qs_heap_min"
      new_drill_heap_max="$def_qs_heap_max"
  fi
  # only change if memory settings are either the QS defaults or Drill defaults
  # if customer has changed them, leave as is
  if [ \( "$conf_drill_heap" = "${def_qs_heap}m" -o "$conf_drill_heap" = "$def_drill_heap" \) -a \
       \( "$conf_drill_mdm" = "${def_qs_mdm}m" -o "$conf_drill_mdm" = "$def_drill_mdm" \) -a \
       \( "$conf_drill_dccs" = "${def_qs_dccs}m" -o "$conf_drill_dccs" = "$def_drill_dccs" \) -a \
       \( "$conf_drill_heap_min" = "${def_qs_heap_min}" -o "$conf_drill_heap_min" = "$def_drill_heap_min" \) -a \
       \( "$conf_drill_heap_max" = "${def_qs_heap_max}" -o "$conf_drill_heap_max" = "$def_drill_heap_max" \) ]; then
      updateMemSettings=1
  fi
  if [ "$updateMemSettings" -eq 1 ]; then
      sed -i -e "s/DRILL_HEAP=[0-9]*[GgMmKkBb]/DRILL_HEAP=$new_drill_heap/" $wardenTmpDrillConf
      sed -i -e "s/DRILL_MAX_DIRECT_MEMORY=[0-9]*[GgMmKkBb]/DRILL_MAX_DIRECT_MEMORY=$new_drill_mdm/" $wardenTmpDrillConf
      sed -i -e "s/DRILLBIT_CODE_CACHE_SIZE=[0-9]*[GgMmKkBb]/DRILLBIT_CODE_CACHE_SIZE=$new_drill_dccs/" $wardenTmpDrillConf
      sed -i -e "s/service.heapsize.min=[0-9]*[GgMmKkBb]*/service.heapsize.min=$new_drill_heap_min/" $wardenTmpDrillConf
      sed -i -e "s/service.heapsize.max=[0-9]*[GgMmKkBb]*/service.heapsize.max=$new_drill_heap_max/" $wardenTmpDrillConf
  fi
}

function CopyDrillbitWardenFile() {
  if [ "$drillBitsNodeOn" -eq 1 ]; then
    if [ "$wardenConfFileInstalled" -eq 0 ]; then
      if ! [ -d "${MAPR_CONF_DIR}" ]; then
        mkdir -p "${MAPR_CONF_DIR}"
      fi
    fi
    cp "$wardenTmpDrillConf" "$wardenInstalledDrillConf"
  fi
}

#
# Create a REST API session with Drill Webserver
#
function LoginToRestApi() {
    local loginUrl="https://localhost:8047/j_security_check"

    # Save the login page result in loginResult. We will use that result to see if it was successful login or not
    local loginResult=`wget -qO- -t 10 --retry-connrefused --save-cookies=$httpsCookieFile --keep-session-cookies --post-data "j_username=$httpUserName&j_password=$httpUserPwd" --no-check-certificate $loginUrl` 

    # Login page will always display the username
    httpsLoginResult=`echo $loginResult | grep -i -c "$username"`
}

#
# Function to validate if REST API call was successfull or not
#
function ValidateHttpOutput() {
    eval httpCmdOutput="$1"
    if echo $httpCmdOutput | egrep -i -q "success" ;then
        logInfo "Successfully updated $2 storage plugin"
    else
        logInfo "ERROR: Failed to update $2 storage plugin"
    fi
}

#
# Function to update HBase Storage Plugin for Drill
#
function UpdateHbasePlugin() {

    local useHttps=$2
    local wgetCmd=""
    local zkHosts=`echo $zkNodesList | sed -e "s/:5181//g"`

    if [ "$useHttps" == "true" ]
    then
        wgetCmd="wget -qO- -t 10 --retry-connrefused --load-cookies $httpsCookieFile --no-check-certificate $1/hbase.json"
    else
        wgetCmd="wget -qO- -t 10 --retry-connrefused $1/hbase.json"
    fi

    # No harm in retrying here as well since if connection is successful in first try then it will not retry
    local hbasePlug=`$wgetCmd`
    local httpOutput=""

    # Update HBase storage plugin. If the plugin is in disabled format then enable it and update the zookeeper related information
    # Else leave the Hbase plugin as is
    if echo $hbasePlug | grep -i -q "\"enabled\" : false"; then
        httpOutput=`$wgetCmd --post-data="{ \"name\": \"hbase\", \"config\": { \"type\" : \"hbase\", \"enabled\" : true, \"config\" : { \"hbase.zookeeper.quorum\" : \"$zkHosts\", \"hbase.zookeeper.property.clientPort\" : \"5181\" }}}" --header="Content-Type: application/json" --header="Accept: application/json"`
    fi
    
    ValidateHttpOutput "\${httpOutput}" "hbase"
}

#
# Function to update Hive Storage Plugin for Drill
#
function UpdateHivePlugin() {

    local useHttps=$2
    local wgetCmd=""
    local httpOutput=""
    local currentSaslValue=""
    local finalSaslValue=""
    
    if [ "$useHttps" == "true" ]
    then
        currentSaslValue="false"
        finalSaslValue=`echo $useHttps`
        wgetCmd="wget -qO- -t 10 --retry-connrefused --load-cookies $httpsCookieFile --no-check-certificate $1/hive.json"
    else
        currentSaslValue="true"
        finalSaslValue=`echo $useHttps`
        wgetCmd="wget -qO- -t 10 --retry-connrefused $1/hive.json"
    fi

    # No harm in retrying here as well since if connection is successful in first try then it will not retry
    local hivePlug=`$wgetCmd`
    
    # Check if plugin is in disabled state
    if echo $hivePlug | egrep -i -q "\"config\" : null|\"enabled\" : false"; then
        if [ -n "$hiveMetastoreHost" ]; then
            if [ -n "$hiveMetastoreDbHost" ]; then
                httpOutput=`$wgetCmd --post-data="{\"name\": \"hive\", \"config\": { \"type\": \"hive\", \"enabled\": true, \"configProps\": {\"hive.metastore.uris\": \"thrift://$hiveMetastoreHost:9083\", \"javax.jdo.option.ConnectionURL\": \"jdbc:mysql:;databaseName=../sample-data/drill_hive_db;create=true\", \"hive.metastore.warehouse.dir\": \"/tmp/drill_hive_wh\", \"fs.default.name\": \"file:///\", \"hive.metastore.sasl.enabled\": \"$finalSaslValue\", \"datanucleus.schema.autoCreateAll\" : \"true\" }}}" --header="Content-Type: application/json" --header="Accept: application/json"`
            else
                httpOutput=`$wgetCmd --post-data="{ \"name\": \"hive\", \"config\": { \"type\": \"hive\", \"enabled\": true, \"configProps\": {\"hive.metastore.uris\": \"thrift://$hiveMetastoreHost:9083\", \"hive.metastore.warehouse.dir\": \"/tmp/drill_hive_wh\", \"fs.default.name\": \"file:///\", \"hive.metastore.sasl.enabled\": \"$finalSaslValue\", \"datanucleus.schema.autoCreateAll\" : \"true\" }}}" --header="Content-Type: application/json" --header="Accept: application/json"`
            fi
        elif echo $hivePlug | egrep -i -q "\"config\" : null"; then
            # Hive Storage Plugin is missing and Metastore information is not provided so leave it as is
	    logInfo "INFO: Hive Storage Plugin configuration is missing. Not updating"
        else
            # HiveMetastore Host is not specified so let's only update the security flag and keep the plugin disabled
	    local updatedHivePlug=`echo $hivePlug | sed -e "s/\(.*sasl.enabled\"[ ]*\:[ ]*\)\"$currentSaslValue\"\(.*\)/\1 \"$finalSaslValue\"\2/g"`
	    updatedHivePlug=`echo $updatedHivePlug | sed -e 's/\"/\\\"/g'`
	    updatedHivePlug=\"$updatedHivePlug\"
	    local httpCmd=`echo $wgetCmd --post-data=$updatedHivePlug --header=\"Content-Type: application/json\" --header=\"Accept: application/json\"`
	    httpOutput=`bash -c "$httpCmd"`
        fi
    else
        # Hive plugin is in enabled state. We should just update the security config
	local updatedHivePlug=`echo $hivePlug | sed -e "s/\(.*sasl.enabled\"[ ]*\:[ ]*\)\"$currentSaslValue\"\(.*\)/\1 \"$finalSaslValue\"\2/g"`
	updatedHivePlug=`echo $updatedHivePlug | sed -e 's/\"/\\\"/g'`
	updatedHivePlug=\"$updatedHivePlug\"
	local httpCmd=`echo $wgetCmd --post-data=$updatedHivePlug --header=\"Content-Type: application/json\" --header=\"Accept: application/json\"`
	httpOutput=`bash -c "$httpCmd"`
    fi

    ValidateHttpOutput "\${httpOutput}" "hive"
}

#
# Update Storage plugin for secure cluster where we use https
#
function UpdateStoragePluginConfig_HTTPS() {
    local url="https://localhost:$httpPort/storage"

    LoginToRestApi

    # Login to Drill using REST API and check if it was successful
    if [ $httpsLoginResult -eq "1" ]
    then
        # Update Hbase plugin
        UpdateHbasePlugin $url "true" 

        # Update Hive Plugin
        UpdateHivePlugin $url "true"
    else
        logInfo "ERROR: Failed to connect to Drillbit webserver in secure mode to update storage plugin"
    fi

    # Everything is completed so let's delete the cookie file
    rm -rf $httpsCookieFile
}

#
# Update Storage plugin for unsecure cluster where we use http
#
function UpdateStoragePluginConfig_HTTP() {
    local url="http://localhost:$httpPort/storage"

    # No need to login here since Drillbit is running in unsecure mode
    # Update Hbase storage plugin
    UpdateHbasePlugin $url "false"

    # Update Hive storage plugin
    UpdateHivePlugin $url "false"
}

# Main

#sets MAPR_USER/MAPR_GROUP/logfile
initCfgEnv

#get opts
drill_usage="usage: $0 [-EC <common opts>] [-R] [--secure|--unsecure|--customSecure ] [-qsOnly] [-qsDrill] [-drillOnly] [-hiveMetastoreHost <metastoreHost> | -hiveMetastoreDbHost <metastoreDBHost>] [-username <user>] [-password <passwd>] [-httpPort <port>]"
if [ ${#} -gt 1 ] ; then
    OPTS=`getopt -a -o h -l EC: -l R -l secure -l unsecure -l customSecure -l qsOnly -l qsDrill -l drillOnly -l hiveMetastoreHost: -l hiveMetastoreDbHost: -l username: -l password: -l httpPort: -- "$@"`
    if [ $? != 0 ] ; then
        echo ${drill_usage}
        exit 2
    fi
    eval set -- "$OPTS"
    
    for i ; do
        case "$i" in
            --EC)
                #Parse Common options
                #Ingore ones we don't care about
                ecOpts=($2)
                shift 2
                restOpts="$@"
                eval set -- "${ecOpts[@]} --"
                for j in "$@" ; do
                    case "$j" in
                        --OT|-OT)
                            OTNodesList="$2"
                            shift 2;;
                        --R)
                            DRILL_CONF_ASSUME_RUNNING_CORE=1
                            shift 1
                            ;;
                        --) shift
                            break;;
                        *)
                            #echo "Ignoring common option $j"
                            shift 1;;
                    esac
                done
                shift 2
                eval set -- "$restOpts"
                ;;
            --R)
                DRILL_CONF_ASSUME_RUNNING_CORE=1
                shift 1
                ;;
            --customSecure)
                drill_secure="custom";
                shift 1;;
            -h)
                echo ${drill_usage}
                exit 2
                ;;
            --drillOnly)
                drill_only=1;
                shift 1;;
            --qsDrill)
                drill_qs_and_drill=1;
                shift 1;;
            --qsOnly)
                drill_qs_only=1;
                shift 1;;
            --secure)
                drill_secure="true";
                shift 1;;
            --unsecure)
                drill_secure="false";
                shift 1;;
            --hiveMetastoreHost)
                shift 1
                hiveMetastoreHost=$1
                shift 1;;
            --hiveMetastoreDbHost)
                shift 1
                hiveMetastoreDbHost=$1
                shift 1;;
            --username)
                shift 1
                httpUserName=$1
                shift 1;;
            --password)
                shift 1
                httpUserPwd=$1
                shift 1;;
            --httpPort)
                shift 1
                httpPort=$1
                shift 1;;
            --)
                shift
                break;;
        esac
    done
else
    echo "${drill_usage}"
    exit 2
fi

zkNodesList=$(getZKServers)
if [ $? -ne 0 -o -z "$zkNodesList" ]; then
    logInfo "WARNING: Skipping Drill Bits Role configuration... zkNodesList not set"
    exit
fi

echo "OTNodesList: $OTNodesList"
clusterName=$(getClusterName)
DRILL_VER=$(GetDrillVersion)
ConfigureDrillBitsRole

if [ "$drill_secure" == "true" ]
then
    EnableSecurity
elif [ "$drill_secure" == "false" ]
then
    DisableSecurity
else
    # Called with --customSecure flag. Do nothing, preserve the configuration as is
fi

# Restart Drillbit by copying the Drillbit warden file
CopyDrillbitWardenFile

# Update StoragePlugin Configuration
#if [ "$drill_secure" == "true" ]
#then
#    UpdateStoragePluginConfig_HTTPS
#elif [ "$drill_secure" == "false" ]
#then
#    UpdateStoragePluginConfig_HTTP
#else
#    # Called with --customSecure flag. Do nothing, preserve the configuration as is
#fi
