#!/bin/bash
#set -x
#Globals
MAPR_HOME="${MAPR_HOME:-__PREFIX__}"
drillHome="${DRILL_HOME:=__INSTALL__}"
drillBase="${MAPR_HOME}/drill"
drill_secure="false"
drill_qs_only=0
drill_only=0
drill_qs_and_drill=0
zkNodesList=""
drillBitsNodeOn=0
notConfiguredYet="${drillHome}/conf/.not_configured_yet"
wardenTmpDrillConf="/tmp/warden.drill-bits.conf"
wardenInstalledDrillConf="warden.drill-bits.conf"
wardenConfFileInstalled=0
httpUserName=""
httpUserPwd=""
httpPort="8047"
httpsLoginResult=""
httpsCookieFile="/tmp/installer_cookie.txt"
DRILL_CONF_ASSUME_RUNNING_CORE=0
drillLogPrefix="Drill:"

# Storage Plugin related parameters
hiveMetastoreHost=$HIVEMETASTORE_HOST
hiveMetastoreDbHost=$HIVEMETASTORE_DATABASE_HOST

# Security related parameters
drillClientAuthMech="[\"MAPRSASL\", \"PLAIN\"]"
drillBitAuthMech="MAPRSASL"
drillAuthPackage="org.apache.drill.exec.rpc.user.security"
drillAuthImpl="pam4j"
drillPamProfiles="[\"sudo\", \"login\"]"

maprPamProfile="mapr-admin"
pamLocation="/etc/pam.d/"

# Security configuration prefixes
impersonationPrefix=impersonation
securityUserPrefix=security.user
securityUserAuthPrefix=$securityUserPrefix.auth
securityBitPrefix=security.bit
httpsPrefix=http.ssl_enabled
systemOptionPrefix=options
useHadoopSSL=ssl.useHadoopConfig
hadoopVer=`cat ${MAPR_HOME}/hadoop/hadoopversion`
hadoopHome="${MAPR_HOME}/hadoop/hadoop-$hadoopVer"
hadoopSSLConfig=$hadoopHome/etc/hadoop
hadoopSSLFileName=ssl-server.xml

if [ -e "${MAPR_HOME}/server/common-ecosystem.sh" ]; then
    . "${MAPR_HOME}/server/common-ecosystem.sh"
else
   echo "Failed to source common-ecosystem.sh"
   exit 0
fi


function GetDrillVersion() {
  local ver=""
  if [ -f "${drillBase}/drillversion" ]; then
    ver=$(cat "${drillBase}/drillversion")
  else
    ver=`ls -t ${drillBase} | sed 's/^.*drill-//' | head -1 | awk '{print $1}'`
  fi
  echo "$ver"
}

function ConfigureRunUserForDrill() {
  local DRILL_DIR CURR_USER CURR_GROUP
  DRILL_DIR=$1
  if [ ! -d $DRILL_DIR/conf ]; then
    return
  fi

  CURR_USER=$2
  CURR_GROUP=$3
  chown -R $CURR_USER:$CURR_GROUP "${DRILL_DIR}/conf" >> $logFile 2>&1
}

function getMemValueFromWardenFile() {
    local rc
    local wFile=$1
    local wIdent=$2
    local wVariable=$3
    local RES=$(cat $wFile | grep "^$wIdent" | sed -e 's/.*'"$wVariable="'\([0-9]*[GgMmKk]*\).*$/\1/')
    rc=$?
    # Translating %age to value in terms of system memory
    if [[ "$RES" = *% ]]; then
      local totalSysMem=`cat /proc/meminfo | grep MemTotal | awk '{print $2}'`
      let RES=${RES:0:${#RES}-1}*$totalSysMem/100
    fi
    echo "$RES"
    return $rc
}

function roundUpToNextGB() {
    local valueInMB=$1
    if [ -z "$valueInMB" ]; then echo 0; return 0; fi
    let valueInGB=$valueInMB/1024
    let deltaInMB=$valueInMB%1024
    if [ $deltaInMB -gt 0 ]; then let valueInGB+=1 ; fi
    let valueInMB=$valueInGB*1024
    echo "$valueInMB"
    return
}

# Copy libjpam out of /opt/mapr/lib directory inside DRILL_HOME/conf/pam directory.
# Then use this path inside distrib-env.sh for java.library.path. We want to avoid
# directly pointing to /opt/mapr/lib since it contains lots of other native libraries
# and jars which can conflict with Drill's version of same libs/jars
# 
# Only called in case when security is enabled
#
# Deprecated/Unused - By default libpam4j will be configured instead of libjpam implementations
# Hence we don't need to copy any native library for that.
function CopyAndUpdatePamLib() {
    logInfo "$drillLogPrefix Starting to copy and update pam module"

    # Copy the pam library from $MAPR_HOME/lib to $DRILL_HOME/conf/pam
    local libPamBase="$MAPR_HOME/lib"
    local libPamSrc="$libPamBase/libjpam.so"
    
    if [ ! -f "$libPamSrc" ]
    then
        logInfo "$drillLogPrefix ERROR: Failed to find libjpam.so under $libPamSrc path"
	return
    fi
    
    # If we are here that means file is present so copy it inside the DRILL_HOME/conf/path
    local drillBaseDir=${drillBase}/drill-${DRILL_VER}
    local drillConfDir="$drillBaseDir/conf"
    local libPamDest="$drillConfDir/pam"

    if [ ! -d "$drillConfDir" ]
    then
        logInfo "$drillLogPrefix ERROR: Failed to find Drill's conf path: $drillConfDir"
	return
    fi

    # Remove and create pam directory
    rm -rf $libPamDest
    mkdir $libPamDest

    if [ $? -ne 0 ]
    then
        logInfo "$drillLogPrefix ERROR: Failed to create pam directory: $libPamDest"
        return
    fi

    # Copy the libjpam.so
    cp -f $libPamSrc $libPamDest
    
    if [ $? -ne 0 ]
    then
        logInfo "$drillLogPrefix ERROR: Failed to copy the libjpam under $libPamDest"
	return
    fi

    # Update distrib-env.sh to correct path of libjpam
    UpdatePamPath true DRILL_JAVA_OPTS $libPamDest
}

# Function to update the PAM library path in distrib-env.sh file based on security is enabled
# or disabled.
#
# In this function ~ is chosen as deliminiter for sed purposefully since pamPath has / in its value
# which causes invalid sed expression to form. 
#
# Parameters:
#           $1 = true/false -- security is enabled/disabled
#           $2 = DRILL_JAVA_OPTS
#           $3 = path of libjpam.so
#
# Deprecated/Unused - By default libpam4j will be configured instead of libjpam implementations
# Hence we don't need to update the java.library.path
function UpdatePamPath() {
     logInfo "$drillLogPrefix Starting to update pam path with security $1 and path $3"

     local javaLibPath="-Djava.library.path="

     # Remove trailing spaces at the end of line
     #sed -i -e "s/\([^ ]*\)\([ ]*$\)/\1/g" $drillDistribEnv

     # Seucrity is disabled hence remove the java.library.path from DRILL_JAVA_OPTS
     if [ "$1" == "false" ]
     then
         sed -i -e "s~ $javaLibPath[^ \"]*~~g" $drillDistribEnv
         return
     fi
     
     # Check if this string exists in distrib-env.sh
     if grep -q "$2.*$javaLibPath" $drillDistribEnv
     then
         # Remove the java.library.path pointing to old pam module path
	 sed -i -e "s~ $javaLibPath[^ \"]*~~g" $drillDistribEnv
     fi
     
     # Remove the last " in the DRILL_JAVA_OPTS
     sed -i -e "s~\(^export $2=.*\)\"$~\1~g" $drillDistribEnv

     # Add java.library.path in DRILL_JAVA_OPTS with pam module path
     sed -i -e "s~\(^export $2=.*\)~\1 $javaLibPath$3\"~g" $drillDistribEnv

     logInfo "$drillLogPrefix Completed updating pam path"
}

# Add or Update zkSaslClientString as true/false for DRILL_JAVA_OPTS
# and SQLLINE_JAVA_OPTS variable inside distrib-env.sh file
#
# Parameters:
#           $1 = DRILL_JAVA_OPTS / SQLLINE_JAVA_OPTS
#           $2 = true/false -- value to change from
#           $3 = true/false -- value to change to
#
function UpdateZookeeperAuth() {
    logInfo "$drillLogPrefix Updating ZookeeperAuth for $1 from $2 to $3"

    # Update the distrib-env.sh file to enable zookeeper security.
    local zkSaslClientString="zookeeper.sasl.client"

    # Check if this string exists in the distrib-env.sh
    if grep -q "$1.*$zkSaslClientString" $drillDistribEnv
    then
        # Update the value to true if it's set to $3 from $2 for all occurences
        sed -i -e "s/\($1.*$zkSaslClientString=\)\($2\)\(.*\)/\1$3\3/g" $drillDistribEnv
    else
        # The zkSaslClientString is absent we need to add one to the end
        # Remove the last " from $1
        sed -i -e "s/\(^export $1=.*\)\"$/\1/g" $drillDistribEnv

        # Now add the zkSaslClientString with " in the end
        sed -i -e "s/\(^export $1=.*\)/\1 -D$zkSaslClientString=$3\"/g" $drillDistribEnv
    fi
}

# Add impersonation config inside the drill-distrib.conf file for Drill
#
# It expects drill-distrib.conf file to be in correct state such that just adding the
# impersonation config string will be done inside drill.exec block by default
#
function AddImpersonationConfig() {
    echo -e "\n  $impersonationPrefix.enabled: true," >> $drillDistribConf
    echo -e "  $impersonationPrefix.max_chained_user_hops: 3," >> $drillDistribConf
}

# Add or Update security configuration inside distrib.conf file for Drill
#
# Enable Authentication/Encryption on Drill
# Enable HTTPS for web channel
#
# Parameter:
#	$1: true/false -- value of enabled: field in security parameters.
#
function EnableSecurityInDistribConf() {
    logInfo "$drillLogPrefix Starting to update security config in $drillDistribConf with security $1"

    # Remove the last closing brace from the config file
    sed -i -e "/}$/d" $drillDistribConf

    # Add config to enable ZK ACL's
    echo -e "\n  zk.apply_secure_acl: $1," >> $drillDistribConf

    # Add client related security config
    AddImpersonationConfig

    echo -e "\n  security.auth.mechanisms: $drillClientAuthMech," >> $drillDistribConf
    echo -e "  $securityUserAuthPrefix.enabled: $1," >> $drillDistribConf
    echo -e "  $securityUserAuthPrefix.packages += \"$drillAuthPackage\"," >> $drillDistribConf
    echo -e "  $securityUserAuthPrefix.impl: \"$drillAuthImpl\"," >> $drillDistribConf

    local maprAdminPamLocation=$pamLocation$maprPamProfile

    # Check if we have mapr-admin pam profile installed inside pam directory and configure
    # pam profile accordingly
    if [ -e $maprAdminPamLocation ]; then
        drillPamProfiles=`echo $drillPamProfiles | sed "s/\]$/, \"$maprPamProfile\"\]"/g`
    fi

    echo -e "  $securityUserAuthPrefix.pam_profiles: $drillPamProfiles," >> $drillDistribConf
    echo -e "  $securityUserPrefix.encryption.sasl.enabled: $1," >> $drillDistribConf
    
    # Add server related security config
    echo -e "\n  $securityBitPrefix.auth.enabled: $1," >> $drillDistribConf
    echo -e "  $securityBitPrefix.auth.mechanism: \"$drillBitAuthMech\"," >> $drillDistribConf
    echo -e "  $securityBitPrefix.encryption.sasl.enabled: $1," >> $drillDistribConf

    # Add config to enable ssl
    echo -e "\n  $httpsPrefix: $1," >> $drillDistribConf
    echo -e "  $useHadoopSSL: $1," >> $drillDistribConf

    # Add closing braces in the end
    echo -e "\n}" >> $drillDistribConf
}

# Delete all the security related configuration added while enabling the security
# for drillbit
# The reason this function is removing last brace, then all the configs and then blank
# lines from end is because it's very though to write something which can remove all the
# blank lines between 2 braces and that too inline. Also I can't just remove all the blank
# lines since there are blank lines in license section of file
function RemoveDrillSecurityConfig() {
    logInfo "$drillLogPrefix Removing all security configs from $drillDistribConf"

    # Remove the last brance
    sed -i -e "/}$/d" $drillDistribConf

    # Remove ZK ACL's related configs
    sed -i -e "/zk.apply_secure_acl.*/d" $drillDistribConf

    # Remove impersonation related configs
    sed -i -e "/$impersonationPrefix.*/d" $drillDistribConf

    # Remove drill client related security configs
    sed -i -e "/security.auth.*/d" $drillDistribConf
    sed -i -e "/$securityUserPrefix.*/d" $drillDistribConf

    # Remove drill bit related security configs
    sed -i -e "/$securityBitPrefix.*/d" $drillDistribConf

    # Remove HTTPS ssl configuration
    sed -i -e "/$httpsPrefix.*/d" $drillDistribConf
    sed -i -e "/$useHadoopSSL.*/d" $drillDistribConf

    # Remove all the blank lines from end
    sed -i -e :a -e '/^\n*$/{$d;N;ba' -e '}' $drillDistribConf

    # Add the last brance
    echo -e "}" >> $drillDistribConf
}

# Enable security for drill
function EnableSecurity() {

    logInfo "$drillLogPrefix Starting to configure with security on"

    # Remove trailing spaces at the end of lines in distrib-env.sh
    sed -i -e "s/\([^ ]*\)\([ ]*$\)/\1/g" $drillDistribEnv

    # Update DRILL_JAVA_OPTS inside distrib-env.sh file
    UpdateZookeeperAuth DRILL_JAVA_OPTS false true
   
    # Update SQQLINE_JAVA_OPTS inside distrib-env.sh file. We don't want to enable
    # Authentication to Zookeeper from DrillClient's
    UpdateZookeeperAuth SQLLINE_JAVA_OPTS true false

    # Copy and update the libjpam.so path inside distrib-env.sh file
    # Commenting below logic since now we will be using pam4j instead of jpam
    # implementation by default. This doesn't require to copy and native libraries
    # But keeping the code here for future purpose if needed to fallback to jpam
    #CopyAndUpdatePamLib

    # Remove drill security config from distrib.conf
    RemoveDrillSecurityConfig
    
    # Update distrib.conf file with security enabled
    EnableSecurityInDistribConf true

    # Add soft link to ssl-server.xml file
    rm -f ${DRILL_CONF_DIR}/$hadoopSSLFileName
    logInfo "$drillLogPrefix Link creation source: $hadoopSSLConfig/$hadoopSSLFileName and dest: ${DRILL_CONF_DIR}/$hadoopSSLFileName"
    ln -s $hadoopSSLConfig/$hadoopSSLFileName ${DRILL_CONF_DIR}/$hadoopSSLFileName
    chown "$MAPR_USER":"$MAPR_GROUP" "${DRILL_CONF_DIR}/$hadoopSSLFileName"

    logInfo "$drillLogPrefix Completed configuring with security on"
}

# Disable security for drill
function DisableSecurity() {
    logInfo "$drillLogPrefix Starting to configure with security off"

    # Remove trailing spaces at the end of lines in distrib-env.sh
    sed -i -e "s/\([^ ]*\)\([ ]*$\)/\1/g" $drillDistribEnv

    # Update DRILL_JAVA_OPTS inside distrib-env.sh file
    UpdateZookeeperAuth DRILL_JAVA_OPTS true false

    # Update SQQLINE_JAVA_OPTS inside distrib-env.sh file. We don't want to enable
    # Authentication to Zookeeper from DrillClient's
    UpdateZookeeperAuth SQLLINE_JAVA_OPTS true false

    # Update the Pam lib path inside distrib-env.sh file
    #UpdatePamPath false DRILL_JAVA_OPTS

    # Remove drill security config from distrib.conf
    RemoveDrillSecurityConfig

    # Remove soft link to ssl-server.xml file
    logInfo "$drillLogPrefix Removing SSL config link: ${DRILL_CONF_DIR}/$hadoopSSLFileName"
    rm -f ${DRILL_CONF_DIR}/$hadoopSSLFileName

    # If configuration is done for Drill as a QS only or for both QS and Drill
    # We will add impersonation setting in insecure case too
    if [ "$drill_qs_only" -eq 1 -o "$drill_qs_and_drill" -eq 1 ]; then
        logInfo "$drillLogPrefix Configured as QS so adding Impersonation Config"

        # Remove the last braces in config file
        sed -i -e "/}$/d" $drillDistribConf

        # Add Impersonation settings
        AddImpersonationConfig

        # Add last braces in config file
        echo -e "\n}" >> $drillDistribConf
    fi
}

function ConfigureDrillBitsRole() {
  local numThreads=4
  local totalMem=""
  local rpcStanza="\n"
  local systemOptionStanza="\n  $systemOptionPrefix.store.parquet.block-size = 268435456"
  local oldClusterId="drillbits1"
  local oldZkConnect="localhost:2181"
  local newClusterId="${clusterName}-drillbits"

  if ! hasRole 'drill-bits' ; then
    logInfo "Skipping Drill Bits Role configuration... Not found"
    return
  else
    drillBitsNodeOn=1
  fi

  logInfo "Configuring Drill Bits Role"
  local DRILL_BASE_DIR=${drillBase}/drill-${DRILL_VER}

  ConfigureRunUserForDrill $DRILL_BASE_DIR $MAPR_USER $MAPR_GROUP

  # DRILL_CONF_DIR used somewhere else so cannot be local
  DRILL_CONF_DIR="${DRILL_BASE_DIR}/conf"
  local drillConf="${DRILL_CONF_DIR}/drill-override.conf"
  local drillDistribQSMemConf="${DRILL_CONF_DIR}/drill-distrib-mem-qs.conf"
  local wardenPkgDrillConf="${DRILL_CONF_DIR}/warden.drill-bits.conf"
  # These are used elsewhere so they cannot be locals
  drillDistribEnv="${DRILL_CONF_DIR}/distrib-env.sh"
  drillDistribConf="${DRILL_CONF_DIR}/drill-distrib.conf"

  if hasRole 'drill-qs' ; then
    if [ $drill_only -eq 1 -o $drill_qs_and_drill -eq 1 ]; then
       rm -f "${MAPR_HOME}/roles/drill-qs"
       drill_qs_only=0
    else
       # -QS option was already used - keep using it
       drill_qs_only=1
    fi
  fi

  if grep -q "cluster-id: \"$oldClusterId\"" $drillConf; then
    # we have a new install with default information - remove it
    sed -i -e "/cluster-id: \"$oldClusterId\"/d" $drillConf
    sed -i -e "/zk.connect: \"$oldZkConnect\"/d" $drillConf
  fi

  # always configure the number of threads required for QS
  if [ -f "$drillDistribQSMemConf" ]; then
    numThreads=$(grep DrillRpcUserThreads "$drillDistribQSMemConf" | cut -d'=' -f2)
    if [ -z "$numThreads" ]; then
      numThreads="4"
    fi
  fi
  rpcStanza="\n  rpc.user.client.threads: $numThreads"

  if [ "$drill_qs_only" -eq 1 ]; then
    touch "${MAPR_HOME}/roles/drill-qs"
    if [ -f "$drillDistribQSMemConf" ]; then
      heapMem=$(grep DRILL_HEAP "$drillDistribQSMemConf" | cut -d'=' -f2)
      directMem=$(grep DRILL_MAX_DIRECT_MEMORY "$drillDistribQSMemConf" | cut -d'=' -f2)
      codeCacheMem=$(grep DRILLBIT_CODE_CACHE_SIZE "$drillDistribQSMemConf" | cut -d'=' -f2)
      let totalMem=$heapMem+$directMem+$codeCacheMem
    fi
  fi

  if [ ! -e "$drillDistribConf" ]; then
    echo -e "drill.exec {\n  cluster-id: \"$newClusterId\",\n  zk.connect: \"$zkNodesList\",${rpcStanza},${systemOptionStanza}\n}" > $drillDistribConf
  else
    sed -i -e "s/\(cluster-id: \).*/\1\"$newClusterId\",/" "$drillDistribConf"
    sed -i -e "s/\(zk.connect: \).*/\1\"$zkNodesList\",/" "$drillDistribConf"
    sed -i -e "s/\(rpc.user.client.threads: \).*/\1\"$numThreads\",/" "$drillDistribConf"
  fi

  # get a warden file to work on
  if [ -e "${MAPR_CONF_CONFD_DIR}/$wardenInstalledDrillConf" ]; then
    cp "${MAPR_CONF_CONFD_DIR}/$wardenInstalledDrillConf" "$wardenTmpDrillConf"
    wardenConfFileInstalled=1
  else
    cp "$wardenPkgDrillConf" "$wardenTmpDrillConf"
  fi
  wardenConfFile="$wardenTmpDrillConf"

  local conf_drill_proc_mem_max=$(getMemValueFromWardenFile $wardenConfFile service.env DRILLBIT_MAX_PROC_MEM)
  local conf_drill_heap_min=$(getMemValueFromWardenFile $wardenConfFile service.heapsize.min service.heapsize.min)
  local conf_drill_heap_max=$(getMemValueFromWardenFile $wardenConfFile service.heapsize.max service.heapsize.max)
  local def_drill_proc_mem_max=$(getMemValueFromWardenFile $wardenPkgDrillConf service.env DRILLBIT_MAX_PROC_MEM)
  local def_drill_heap_min=$(getMemValueFromWardenFile $wardenPkgDrillConf service.heapsize.min service.heapsize.min)
  local def_drill_heap_max=$(getMemValueFromWardenFile $wardenPkgDrillConf service.heapsize.max service.heapsize.max)
  local def_qs_heap=$(getMemValueFromWardenFile $drillDistribQSMemConf 'export DRILL_HEAP' DRILL_HEAP)
  local def_qs_mdm=$(getMemValueFromWardenFile $drillDistribQSMemConf 'export DRILL_MAX_DIRECT_MEMORY' DRILL_MAX_DIRECT_MEMORY)
  local def_qs_dccs=$(getMemValueFromWardenFile $drillDistribQSMemConf 'export DRILLBIT_CODE_CACHE_SIZE' DRILLBIT_CODE_CACHE_SIZE)
  let def_qs_proc_mem_max=$def_qs_heap+$def_qs_mdm+$def_qs_dccs
  local def_qs_heap_min=$def_qs_proc_mem_max
  local def_qs_heap_max=$def_qs_heap_min
  local new_drill_proc_mem_max=""
  local new_drill_heap_min=""
  local new_drill_heap_max=""
  local updateMemSettings=0

  # configure memory requirement
  if [ "$drill_qs_only" -eq 0 ]; then
      new_drill_proc_mem_max=$def_drill_proc_mem_max
      new_drill_heap_min=$def_drill_heap_min
      new_drill_heap_max=$def_drill_heap_max
  else
      new_drill_heap="${def_qs_heap}m"
      new_drill_mdm="${def_qs_mdm}m"
      new_drill_dccs="${def_qs_dccs}m"
      new_drill_proc_mem_max=$(roundUpToNextGB $def_qs_proc_mem_max)"m"
      new_drill_heap_min=$(roundUpToNextGB $def_qs_heap_min)
      new_drill_heap_max=$(roundUpToNextGB $def_qs_heap_max)
  fi

  # only change if memory settings are either the QS defaults or Drill defaults
  # if customer has changed them, leave as is
  # For QS, we only check the conf_drill_proc_mem_max to detect if customer has changed
  if [ \( "$conf_drill_proc_mem_max" = "${def_qs_proc_mem_max}m" -o "$conf_drill_proc_mem_max" = "$def_drill_proc_mem_max" \) -a \
       \( "$conf_drill_heap_min" = "${def_qs_heap_min}" -o "$conf_drill_heap_min" = "$def_drill_heap_min" \) -a \
       \( "$conf_drill_heap_max" = "${def_qs_heap_max}" -o "$conf_drill_heap_max" = "$def_drill_heap_max" \) ]; then
      updateMemSettings=1
  fi

  if [ "$updateMemSettings" -eq 1 ]; then
    #Updating and undo existing file contents if switching between QS and Drill
    if [ "$drill_qs_only" -eq 0 ]; then
      #Delete MaxHeap, Direct and CodeCache params
      sed -i '/DRILL_HEAP/d' $wardenTmpDrillConf
      sed -i '/DRILL_MAX_DIRECT_MEMORY/d' $wardenTmpDrillConf
      sed -i '/DRILLBIT_CODE_CACHE_SIZE/d' $wardenTmpDrillConf
      sed -i -e "s/DRILLBIT_MAX_PROC_MEM=[0-9]*[GgMmKkBb]/DRILLBIT_MAX_PROC_MEM=$new_drill_proc_mem_max/" $wardenTmpDrillConf
    else
      #Place holder for listing missing params
      new_drill_service_env="";
      new_drill_service_env="${new_drill_service_env},DRILL_HEAP=$new_drill_heap"
      new_drill_service_env="${new_drill_service_env},DRILL_MAX_DIRECT_MEMORY=$new_drill_mdm"
      new_drill_service_env="${new_drill_service_env},DRILLBIT_CODE_CACHE_SIZE=$new_drill_dccs"
      #Prepending total (Note: comma already exists)
      new_drill_service_env="DRILLBIT_MAX_PROC_MEM=$new_drill_proc_mem_max${new_drill_service_env}"
      # Substituting all entries in service.env
      sed -i -e "s/service.env=.*/service.env=$new_drill_service_env/" $wardenTmpDrillConf
      sed -i -e "s/service.heapsize.min=[0-9]*[GgMmKkBb]*/service.heapsize.min=$new_drill_heap_min/" $wardenTmpDrillConf
      sed -i -e "s/service.heapsize.max=[0-9]*[GgMmKkBb]*/service.heapsize.max=$new_drill_heap_max/" $wardenTmpDrillConf
    fi
  fi
}

function CopyDrillbitWardenFile() {
  if [ "$drillBitsNodeOn" -eq 1 ]; then
    if [ "$wardenConfFileInstalled" -eq 0 ]; then
      if ! [ -d "${MAPR_CONF_CONFD_DIR}" ]; then
        mkdir -p "${MAPR_CONF_CONFD_DIR}"
      fi
    fi
    if ! grep -q "drill-$DRILL_VER" "$wardenTmpDrillConf" ; then
      sed -i -e "s@/drill-[0-9][0-9]*\.[0-9][0-9]*\.[0-9][0-9]*/@/drill-$DRILL_VER/@" "$wardenTmpDrillConf"
    fi
    cp "$wardenTmpDrillConf" "${MAPR_CONF_CONFD_DIR}/$wardenInstalledDrillConf"
    chown "$MAPR_USER":"$MAPR_GROUP" "${MAPR_CONF_CONFD_DIR}/$wardenInstalledDrillConf"
    rm -f "$wardenTmpDrillConf"
  fi
}

#
# Create a REST API session with Drill Webserver
#
function LoginToRestApi() {
    local loginUrl="https://localhost:8047/j_security_check"

    # Save the login page result in loginResult. We will use that result to see if it was successful login or not
    local loginResult=`wget -qO- -t 10 --retry-connrefused --save-cookies=$httpsCookieFile --keep-session-cookies --post-data "j_username=$httpUserName&j_password=$httpUserPwd" --no-check-certificate $loginUrl` 

    # Login page will always display the username
    httpsLoginResult=`echo $loginResult | grep -i -c "$username"`
}

#
# Function to validate if REST API call was successfull or not
#
function ValidateHttpOutput() {
    eval httpCmdOutput="$1"
    if echo $httpCmdOutput | egrep -i -q "success" ;then
        logInfo "Successfully updated $2 storage plugin"
    else
        logInfo "ERROR: Failed to update $2 storage plugin"
    fi
}

#
# Function to update HBase Storage Plugin for Drill
#
function UpdateHbasePlugin() {

    local useHttps=$2
    local wgetCmd=""
    local zkHosts=`echo $zkNodesList | sed -e "s/:5181//g"`

    if [ "$useHttps" == "true" ]
    then
        wgetCmd="wget -qO- -t 10 --retry-connrefused --load-cookies $httpsCookieFile --no-check-certificate $1/hbase.json"
    else
        wgetCmd="wget -qO- -t 10 --retry-connrefused $1/hbase.json"
    fi

    # No harm in retrying here as well since if connection is successful in first try then it will not retry
    local hbasePlug=`$wgetCmd`
    local httpOutput=""

    # Update HBase storage plugin. If the plugin is in disabled format then enable it and update the zookeeper related information
    # Else leave the Hbase plugin as is
    if echo $hbasePlug | grep -i -q "\"enabled\" : false"; then
        httpOutput=`$wgetCmd --post-data="{ \"name\": \"hbase\", \"config\": { \"type\" : \"hbase\", \"enabled\" : true, \"config\" : { \"hbase.zookeeper.quorum\" : \"$zkHosts\", \"hbase.zookeeper.property.clientPort\" : \"5181\" }}}" --header="Content-Type: application/json" --header="Accept: application/json"`
    fi
    
    ValidateHttpOutput "\${httpOutput}" "hbase"
}

#
# Function to update Hive Storage Plugin for Drill
#
function UpdateHivePlugin() {

    local useHttps=$2
    local wgetCmd=""
    local httpOutput=""
    local currentSaslValue=""
    local finalSaslValue=""
    
    if [ "$useHttps" == "true" ]
    then
        currentSaslValue="false"
        finalSaslValue=`echo $useHttps`
        wgetCmd="wget -qO- -t 10 --retry-connrefused --load-cookies $httpsCookieFile --no-check-certificate $1/hive.json"
    else
        currentSaslValue="true"
        finalSaslValue=`echo $useHttps`
        wgetCmd="wget -qO- -t 10 --retry-connrefused $1/hive.json"
    fi

    # No harm in retrying here as well since if connection is successful in first try then it will not retry
    local hivePlug=`$wgetCmd`
    
    # Check if plugin is in disabled state
    if echo $hivePlug | egrep -i -q "\"config\" : null|\"enabled\" : false"; then
        if [ -n "$hiveMetastoreHost" ]; then
            if [ -n "$hiveMetastoreDbHost" ]; then
                httpOutput=`$wgetCmd --post-data="{\"name\": \"hive\", \"config\": { \"type\": \"hive\", \"enabled\": true, \"configProps\": {\"hive.metastore.uris\": \"thrift://$hiveMetastoreHost:9083\", \"javax.jdo.option.ConnectionURL\": \"jdbc:mysql:;databaseName=../sample-data/drill_hive_db;create=true\", \"hive.metastore.warehouse.dir\": \"/tmp/drill_hive_wh\", \"fs.default.name\": \"file:///\", \"hive.metastore.sasl.enabled\": \"$finalSaslValue\", \"datanucleus.schema.autoCreateAll\" : \"true\" }}}" --header="Content-Type: application/json" --header="Accept: application/json"`
            else
                httpOutput=`$wgetCmd --post-data="{ \"name\": \"hive\", \"config\": { \"type\": \"hive\", \"enabled\": true, \"configProps\": {\"hive.metastore.uris\": \"thrift://$hiveMetastoreHost:9083\", \"hive.metastore.warehouse.dir\": \"/tmp/drill_hive_wh\", \"fs.default.name\": \"file:///\", \"hive.metastore.sasl.enabled\": \"$finalSaslValue\", \"datanucleus.schema.autoCreateAll\" : \"true\" }}}" --header="Content-Type: application/json" --header="Accept: application/json"`
            fi
        elif echo $hivePlug | egrep -i -q "\"config\" : null"; then
            # Hive Storage Plugin is missing and Metastore information is not provided so leave it as is
	    logInfo "INFO: Hive Storage Plugin configuration is missing. Not updating"
        else
            # HiveMetastore Host is not specified so let's only update the security flag and keep the plugin disabled
	    local updatedHivePlug=`echo $hivePlug | sed -e "s/\(.*sasl.enabled\"[ ]*\:[ ]*\)\"$currentSaslValue\"\(.*\)/\1 \"$finalSaslValue\"\2/g"`
	    updatedHivePlug=`echo $updatedHivePlug | sed -e 's/\"/\\\"/g'`
	    updatedHivePlug=\"$updatedHivePlug\"
	    local httpCmd=`echo $wgetCmd --post-data=$updatedHivePlug --header=\"Content-Type: application/json\" --header=\"Accept: application/json\"`
	    httpOutput=`bash -c "$httpCmd"`
        fi
    else
        # Hive plugin is in enabled state. We should just update the security config
	local updatedHivePlug=`echo $hivePlug | sed -e "s/\(.*sasl.enabled\"[ ]*\:[ ]*\)\"$currentSaslValue\"\(.*\)/\1 \"$finalSaslValue\"\2/g"`
	updatedHivePlug=`echo $updatedHivePlug | sed -e 's/\"/\\\"/g'`
	updatedHivePlug=\"$updatedHivePlug\"
	local httpCmd=`echo $wgetCmd --post-data=$updatedHivePlug --header=\"Content-Type: application/json\" --header=\"Accept: application/json\"`
	httpOutput=`bash -c "$httpCmd"`
    fi

    ValidateHttpOutput "\${httpOutput}" "hive"
}

#
# Update Storage plugin for secure cluster where we use https
#
function UpdateStoragePluginConfig_HTTPS() {
    local url="https://localhost:$httpPort/storage"

    LoginToRestApi

    # Login to Drill using REST API and check if it was successful
    if [ $httpsLoginResult -eq "1" ]
    then
        # Update Hbase plugin
        UpdateHbasePlugin $url "true" 

        # Update Hive Plugin
        UpdateHivePlugin $url "true"
    else
        logInfo "ERROR: Failed to connect to Drillbit webserver in secure mode to update storage plugin"
    fi

    # Everything is completed so let's delete the cookie file
    rm -rf $httpsCookieFile
}

#
# Update Storage plugin for unsecure cluster where we use http
#
function UpdateStoragePluginConfig_HTTP() {
    local url="http://localhost:$httpPort/storage"

    # No need to login here since Drillbit is running in unsecure mode
    # Update Hbase storage plugin
    UpdateHbasePlugin $url "false"

    # Update Hive storage plugin
    UpdateHivePlugin $url "false"
}

# Main

#sets MAPR_USER/MAPR_GROUP/logfile
initCfgEnv

#get opts
drill_usage="usage: $0 [-EC <common opts>] [-R] [--secure|--unsecure|--customSecure ] [-qsOnly] [-qsDrill] [-drillOnly] [-hiveMetastoreHost <metastoreHost> | -hiveMetastoreDbHost <metastoreDBHost>] [-username <user>] [-password <passwd>] [-httpPort <port>]"
if [ ${#} -ge 1 ] ; then
    OPTS=`getopt -a -o h -l EC: -l R -l secure -l unsecure -l customSecure -l qsOnly -l qsDrill -l drillOnly -l hiveMetastoreHost: -l hiveMetastoreDbHost: -l username: -l password: -l httpPort: -- "$@"`
    if [ $? != 0 ] ; then
        echo ${drill_usage}
        exit 2
    fi
    eval set -- "$OPTS"
    
    for i ; do
        case "$i" in
            --EC)
                #Parse Common options
                #Ingore ones we don't care about
                ecOpts=($2)
                shift 2
                restOpts="$@"
                eval set -- "${ecOpts[@]} --"
                for j in "$@" ; do
                    case "$j" in
                        --OT|-OT)
                            OTNodesList="$2"
                            shift 2;;
                        --R|-R)
                            DRILL_CONF_ASSUME_RUNNING_CORE=1
                            shift 1
                            ;;
                        --QS|-QS)
                            drill_qs_only=1
                            shift 1
                            ;;
                        --) shift
                            break;;
                        *)
                            #echo "Ignoring common option $j"
                            shift 1;;
                    esac
                done
                shift 2
                eval set -- "$restOpts"
                ;;
            --R)
                DRILL_CONF_ASSUME_RUNNING_CORE=1
                shift 1
                ;;
            --customSecure)
                drill_secure="custom";
                shift 1;;
            -h)
                echo ${drill_usage}
                exit 2
                ;;
            --drillOnly)
                drill_only=1;
                shift 1;;
            --qsDrill)
                drill_qs_and_drill=1;
                shift 1;;
            --qsOnly)
                drill_qs_only=1;
                shift 1;;
            --secure)
                drill_secure="true";
                shift 1;;
            --unsecure)
                drill_secure="false";
                shift 1;;
            --hiveMetastoreHost)
                shift 1
                hiveMetastoreHost=$1
                shift 1;;
            --hiveMetastoreDbHost)
                shift 1
                hiveMetastoreDbHost=$1
                shift 1;;
            --username)
                shift 1
                httpUserName=$1
                shift 1;;
            --password)
                shift 1
                httpUserPwd=$1
                shift 1;;
            --httpPort)
                shift 1
                httpPort=$1
                shift 1;;
            --)
                shift
                break;;
        esac
    done
fi

zkNodesList=$(getZKServers)
if [ $? -ne 0 -o -z "$zkNodesList" ]; then
    logInfo "WARNING: Skipping Drill Bits Role configuration... zkNodesList not set"
    exit
fi

echo "OTNodesList: $OTNodesList"
clusterName=$(getClusterName)
DRILL_VER=$(GetDrillVersion)
ConfigureDrillBitsRole

if [ "$drill_secure" == "true" ]
then
    EnableSecurity
elif [ "$drill_secure" == "false" ]
then
    DisableSecurity
else
    # Called with --customSecure flag. Do nothing, preserve the configuration as is
    # except in the case were we are added to a node in customSecure mode
    # create initial configuration as if are running in MapR secure mode.
    # customer will need to edit to match own configuration and restart drill
    if [ -f "$notConfiguredYet" ]; then
        EnableSecurity
    fi
fi

# Restart Drillbit by copying the Drillbit warden file
CopyDrillbitWardenFile

# Update StoragePlugin Configuration
#if [ "$drill_secure" == "true" ]
#then
#    UpdateStoragePluginConfig_HTTPS
#elif [ "$drill_secure" == "false" ]
#then
#    UpdateStoragePluginConfig_HTTP
#else
#    # Called with --customSecure flag. Do nothing, preserve the configuration as is
#fi

# remove state flag indicating fresh install
if [ -f "$notConfiguredYet" ]; then
    rm -f "$notConfiguredYet"
fi
